/* SCRED.CPP - editor of ZX-Spectrum images in SCR-format for DOS (386+)

Copyright (c) 2001,2002,2021 Alexander Shabarshin <me@shaos.net>

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

// Aug 20, 2001 - Started working on it using WATCOM-C/C++ 10.0
// Version 1.0 (Aug 25, 2001)
// Version 1.1 (Jul 05, 2002)
// Version 1.2 (Oct 20, 2002)
// Version 1.3 (Apr 1, 2021) - added 5x7 font and running text
// Compilable by Open Watcom C/C++ v1.9

#include <graph.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>
#include <time.h>
#include <i86.h>
#include "my_mouse.hpp"
#include "font5x7.hpp"

#define DDX 640
#define DDY 480

Mouse um(U_TEXT); // it's text to simplify positioning

void MouseInit()
{
  um.Show();
  um.Sensitive(16,16);
}

int zx(int i)
{
  int o = 0;
  switch(i&0x07)
  {
    case 0: o = 0; break;
    case 1: o = 1; break;
    case 2: o = 4; break;
    case 3: o = 5; break;
    case 4: o = 2; break;
    case 5: o = 3; break;
    case 6: o = 6; break;
    case 7: o = 7; break;
  }
  return o;
}

unsigned char *image;
unsigned char *attr;
int saved = 1;

void DrawPalette(int k,int cs=0,int cb=7)
{
  int x0 = 528;
  int y0 = 160;
  int yy;
  int i,j,cola;
  _setcolor(0);
  _rectangle(0,x0-9,y0,x0-2,y0+128);
  _rectangle(0,x0+55,y0,x0+62,y0+128);
  for(i=0;i<2;i++)
  {
    yy = y0;
    for(j=0;j<8;j++)
    {
       cola = k<<3;
       if(!j) cola = 0;
       _setcolor(zx(j)+cola);
       _rectangle(0,x0,yy,x0+46,yy+15);
       yy += 16;
    }
    _setcolor(14);
    _moveto(x0-1,y0-1);
    _lineto(x0+47,y0-1);
    _lineto(x0+47,y0+128);
    _lineto(x0-1,y0+128);
    _lineto(x0-1,y0-1);
    x0 += 64;
  }
  j = (cb+10)<<4;
  _moveto(519,j);
  _lineto(527,j+8);
  _lineto(519,j+16);
  j = (cs+10)<<4;
  _moveto(583,j);
  _lineto(591,j+8);
  _lineto(583,j+16);
}

void DrawZnako(int x,int y,int f) // f=0 - b/w, f=1 - screen, f=2 - big
{
  #define SZ  16
  #define SZ1 15
  #define SZB 4
  int x0 = 512;
  if(f<2) x0 = 0;
  int y0 = 16;
  int at = x+(y<<5);
  int i,j;
  int ii,jj;
  int im = x+(y<<8);
  int k = attr[at];
  int cs = k&0x07;
  int cb = (k&0x38)>>3;
  if(f==2) DrawPalette((k&0x40)?1:0,cs,cb);
  cs = zx(cs);
  cb = zx(cb);
  if(k&0x40)
  {
     if(cs) cs |= 0x08;
     if(cb) cb |= 0x08;
  }
  for(j=0;j<8;j++)
  {
     k = image[im];
     for(i=0;i<8;i++)
     {
        if((k<<i)&0x80)
        {
          if(f) _setcolor(cs);
          else  _setcolor(0);
        }
        else
        {
          if(f) _setcolor(cb);
          else  _setcolor(15);
        }
        if(f==2)
        {
          ii = x0+(i<<SZB);
          jj = y0+(j<<SZB);
          _rectangle(0,ii,jj,ii+SZ1,jj+SZ1);
        }
        if(f<2)
        {
          ii = x0+(x<<4)+i+i;
          jj = y0+(y<<4)+j+j;
          _setpixel(ii,jj);
          _setpixel(ii,jj+1);
          _setpixel(ii+1,jj);
          _setpixel(ii+1,jj+1);
        }
     }
     im += 32;
  }
}

void DrawRecta(int ix,int iy)
{
  int x = ix<<4;
  int y = (iy+1)<<4;
  int i,j,k;
  for(i=0;i<16;i+=2)
  {
     _setcolor(1);
     _setpixel(x+i,y);
     _setpixel(x+i,y+15);
     _setcolor(14);
     _setpixel(x+i+1,y);
     _setpixel(x+i+1,y+15);
  }
  for(j=0;j<16;j+=2)
  {
     _setcolor(1);
     _setpixel(x,y+j);
     _setpixel(x+15,y+j);
     _setcolor(14);
     _setpixel(x,y+j+1);
     _setpixel(x+15,y+j+1);
  }
}

void ChangePix(int zx, int zy, int x, int y, int c)
{
  int i;
  int mas = 0x80;
  int mai;
  int im = zx+(((zy<<3)+y)<<5);
  for(i=0;i<x;i++) mas>>=1;
  mai = ~mas;
  if(image[im]&mas)
  {
     if(c==0) image[im] &= mai;
  }
  else
  {
     if(c==1) image[im] |= mas;
  }
}

void InvZnako(int zx, int zy)
{
  int i,j,k;
  int im;
  for(j=0;j<8;j++)
  {
    im = zx+(((zy<<3)+j)<<5);
    k = image[im];
    image[im] = ~k;
  }
}

unsigned char ImageBuf[9];

void CopyZnako(int zx, int zy)
{
  int i,j,k;
  int im;
  for(j=0;j<8;j++)
  {
    im = zx+(((zy<<3)+j)<<5);
    ImageBuf[j] = image[im];
  }
  ImageBuf[8] = attr[zx+(zy<<5)];
}

void PasteZnako(int zx, int zy)
{
  int i,j,k;
  int im;
  for(j=0;j<8;j++)
  {
    im = zx+(((zy<<3)+j)<<5);
    image[im] = ImageBuf[j];
  }
  attr[zx+(zy<<5)] = ImageBuf[8];
}

void OutChar(int x, int y, char c, int s)
{
  if(c<0 || c>=240) return;
  int co = _getcolor();
  _setcolor(0);
  _rectangle(0,x,y,x+6*s-1,y+7*s-1);
  _setcolor(co);
  int yy = y;
  for(int j=0;j<7;j++)
  {
     int l = font5x7[c-32][j];
     int m = 0x80;
     int xx = x;
     for(int i=0;i<5;i++)
     {
        if(l&m)
        {
          if(xx>=0 && xx<DDX && yy>=0 && yy<DDY) _setpixel(xx,yy);
        }
        m >>= 1;
        xx += s;
     }
     yy += s;
  }
}

int OutText(int x, int y, char* text, int zoom=1)
{
  int st = 6*zoom;
  for(int i=0;i<strlen(text);i++)
  {
     OutChar(x,y,text[i],zoom);
     x += st;
     if(x>=DDX) return 0;
  }
  return 1;
}

#define LEDY 427
#define MSGX 539
#define MSGY 464

char *copyr = "SCR-EDitor v1.3 Copyright (c) 2001,2002,2021 "
              "Alexander \"Shaos\" Shabarshin <me@shaos.net> "
              "http://nedoPC.org";

int main(int argc, char **argv)
{
 char str[256];
 int i,j,ii,jj,i1,j1,j2,k,cs,cb,ci;
 int sx,sy;
 FILE *f;
 int szi = 256*192/8;
 int sza = 32*24;
 image = new unsigned char[szi];
 if(image==NULL) return 0;
 attr = new unsigned char[sza];
 if(attr==NULL) return 0;
 _setvideomode(18);
 _setcolor(0);
 _rectangle(0,0,0,DDX-1,DDY-1);
 _setcolor(7);
 OutText(3,4,copyr);
 if(argc>1)
 {
   f = fopen(argv[1],"rb");
   if(f==NULL)
   {
      _setvideomode(-1);
      return 0;
   }
   for(i=0;i<szi;i++)
   {
      ii = i&0x1F;
      j = i>>5;
      j1 = j&0x07;
      j2 = (j&0x38)>>3;
      jj = (j&0xC0)|(j1<<3)|j2;
      image[ii|(jj<<5)] = fgetc(f);
   }
   for(i=0;i<sza;i++)
   {
      attr[i] = fgetc(f);
   }
   fclose(f);
   for(j=0;j<192;j++){
   for(i=0;i<256;i+=8){
      k = attr[((j>>3)<<5)|(i>>3)];
      cs = zx(k&0x07);
      cb = zx((k&0x38)>>3);
      if(k&0x40)
      {
         if(cs) cs |= 0x08;
         if(cb) cb |= 0x08;
      }
      k = image[(i>>3)|(j<<5)];
      for(ii=0;ii<8;ii++)
      {
         if((k<<ii)&0x80)
            _setcolor(cs);
         else
            _setcolor(cb);
         i1 = (i+ii)<<1;
         j1 = (j<<1)+16;
         _setpixel(i1,j1);
         _setpixel(i1,j1+1);
         _setpixel(i1+1,j1);
         _setpixel(i1+1,j1+1);
      }
   }}
 }
 else
 {
   for(i=0;i<sza;i++) attr[i] = 0x38;
   for(i=0;i<szi;i++) image[i] = 0x00;
   for(j=0;j<24;j++){
   for(i=0;i<32;i++){
      DrawZnako(i,j,1);
   }}
 }
 _setcolor(15);
 OutText(515,80,"<- Click there first");
 _setcolor(14);
 OutText(537,149,"PAPER");
 OutText(607,149,"INK");
 _rectangle(2,528,296,638,328);
 OutText(549,307,"BRIGHT",2);
 _rectangle(2,528,328,638,360);
 OutText(549,339,"INVERT",2);
 _rectangle(2,528,360,638,392);
 OutText(561,371,"COPY",2);
 _rectangle(2,528,392,638,424);
 OutText(555,403,"PASTE",2);
 _rectangle(2,528,424,638,456);
 OutText(561,435,"SAVE",2);
 _setcolor(12);
 OutText(30,LEDY,"S",5);
 _setcolor(14);
 OutText(60,LEDY,"C",5);
 _setcolor(10);
 OutText(90,LEDY,"R",5);
 _setcolor(9);
 OutText(120,LEDY,"E",5);
 _setcolor(15);
 OutText(150,LEDY,"D by Shaos >",5);
 DrawPalette(0);
 MouseInit();
 char *text = copyr;
 if(argc>2) strcpy(str,argv[2]);
 else strcpy(str,"scred.cpp");
 f = fopen(str,"rb");
 if(f!=NULL)
 {
   fseek(f,0,SEEK_END);
   k = ftell(f);
   fseek(f,0,SEEK_SET);
   text = (char*)malloc(k);
   if(text==NULL)
   {
     text = copyr;
     fclose(f);
     f = NULL;
   }
   else *text = 0;
   while(f)
   {
     str[0] = 0xB1;
     fgets(&str[1],255,f);
     str[255] = 0;
     char *po = strrchr(str,'\r');
     if(po!=NULL) *po=0;
     if(str[1]) strcat(text,str);
     if(feof(f)) break;
   }
   fclose(f);
 }
 char *buf = (char*)malloc(_imagesize(1,LEDY,512,LEDY+34));
 int ix = 0;
 int iy = 0;
 int counter = 0;
 int scroll = -1;
 int ptr = 0;
 int offset = 6;
 int key,letter;
 while(1)
 {
   if(scroll>0)
   {
      _getimage(1,LEDY,512,LEDY+34,buf);
      _putimage(0,LEDY,buf,_GPSET);
      if(++counter==5)
      {
         counter = 0;
         if(offset==6)
         {
            offset = 0;
            int oldl = letter;
            letter = text[ptr++];
            if(letter==0)
            {
               ptr = 0;
               letter = 0xB1;
            }
            if(letter==9) letter=' ';
            if(letter<=32 && oldl==' ') letter = text[ptr++];
         }
         int mm = 0x80>>offset;
         int kk = letter-32;
         _setcolor(7);
         if(font5x7[kk][0] & mm) _setpixel(511,LEDY);
         if(font5x7[kk][1] & mm) _setpixel(511,LEDY+5);
         if(font5x7[kk][2] & mm) _setpixel(511,LEDY+10);
         if(font5x7[kk][3] & mm) _setpixel(511,LEDY+15);
         if(font5x7[kk][4] & mm) _setpixel(511,LEDY+20);
         if(font5x7[kk][5] & mm) _setpixel(511,LEDY+25);
         if(font5x7[kk][6] & mm) _setpixel(511,LEDY+30);
         offset++;
      }
   }
   key = 0;
   if(kbhit())
   {
      key = getch();
      if(key==27) break;
      key = toupper(key);
   }
   um.Position();
   sx = um.GetX()<<3;
   sy = um.GetY()<<3;
   if(um.GetButton(U_RIGHT))
   {
      um.Hide();
      if(sx>=512&&sx<640&&sy>=16&&sy<144)
      {
         ii = (sx-512)>>4;
         jj = (sy-16)>>4;
         ChangePix(ix,iy,ii,jj,0);
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(100);
      }
      um.Show();
   }
   if(um.GetButton(U_LEFT)||key)
   {
      if(scroll<0) scroll=1;
      um.Hide();
      if(!key&&sx>=0&&sx<512&&sy>=400&&sy<480)
      {
         if(scroll==0) scroll=1;
         else if(scroll==1) scroll=0;
         delay(300);
      }
      if(!key&&sx>=0&&sx<512&&sy>=16&&sy<400)
      {
         DrawZnako(ix,iy,1);
         ix = (sx>>4);
         iy = (sy>>4)-1;
         DrawRecta(ix,iy);
         DrawZnako(ix,iy,2);
      }
      if(!key&&sx>=512&&sx<640&&sy>=16&&sy<144)
      {
         ii = (sx-512)>>4;
         jj = (sy-16)>>4;
         ChangePix(ix,iy,ii,jj,1);
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(100);
      }
      if(!key&&sx>=512&&sx<640&&sy>=160&&sy<288)
      {
         jj = (sy-160)>>4;
         k = attr[ix+(iy<<5)];
         if(sx-512 < 64)
            k = (k&0xC7)|(jj<<3);
         else
            k = (k&0xF8)|jj;
         attr[ix+(iy<<5)] = k;
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(200);
      }
      if(key=='B'||(sx>=528&&sx<640&&sy>=296&&sy<328))
      {
         _setcolor(15);
         _rectangle(2,530,298,636,326);
         k = attr[ix+(iy<<5)];
         if(k&0x40)
            k = k&0x3F;
         else
            k = k|0x40;
         attr[ix+(iy<<5)] = k;
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(200);
         _setcolor(0);
         _rectangle(2,530,298,636,326);
      }
      if(key=='I'||(sx>=528&&sx<640&&sy>=328&&sy<360))
      {
         _setcolor(15);
         _rectangle(2,530,330,636,358);
         InvZnako(ix,iy);
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(200);
         _setcolor(0);
         _rectangle(2,530,330,636,358);
      }
      if(key=='C'||(sx>=528&&sx<640&&sy>=360&&sy<392))
      {
         _setcolor(15);
         _rectangle(2,530,362,636,390);
         CopyZnako(ix,iy);
         delay(200);
         _setcolor(0);
         _rectangle(2,530,362,636,390);
      }
      if(key=='P'||(sx>=528&&sx<640&&sy>=392&&sy<424))
      {
         _setcolor(15);
         _rectangle(2,530,394,636,422);
         PasteZnako(ix,iy);
         DrawZnako(ix,iy,2);
         DrawZnako(ix,iy,1);
         DrawRecta(ix,iy);
         delay(200);
         _setcolor(0);
         _rectangle(2,530,394,636,422);
      }
      if(key=='S'||(sx>=528&&sx<640&&sy>=424&&sy<456))
      {
         _setcolor(15);
         _rectangle(2,530,426,636,454);
         sprintf(str,"%4.4X.scr",saved++);
         f = fopen(str,"wb");
         if(f!=NULL)
         {
           for(i=0;i<szi;i++)
           {
              ii = i&0x1F;
              j = i>>5;
              j1 = j&0x07;
              j2 = (j&0x38)>>3;
              jj = (j&0xC0)|(j1<<3)|j2;
              fputc(image[ii|(jj<<5)],f);
           }
           for(i=0;i<sza;i++)
           {
              fputc(attr[i],f);
           }
           fclose(f);
           strcat(str," saved!");
           _setcolor(7);
           OutText(MSGX,MSGY,str);
           delay(200);
           _setcolor(0);
           _rectangle(2,530,426,636,454);
         }
      }
      if(key=='W'||(sx>=0&&sx<640&&sy>=0&&sy<16))
      {
         _setcolor(7);
         OutText(MSGX,MSGY,"Temporarily B/W");
         for(j=0;j<24;j++){
         for(i=0;i<32;i++){
            DrawZnako(i,j,0);
         }}
         delay(1000);
         _setcolor(7);
         OutText(MSGX,MSGY,"     Color     ");
         for(j=0;j<24;j++){
         for(i=0;i<32;i++){
            DrawZnako(i,j,1);
         }}
      }
      if(key=='Z') // nedoPC SDK sprite format
      {
        sprintf(str,"%4.4X.spr",saved++);
        f = fopen(str,"wt");
        if(f!=NULL)
        {
           for(j=0;j<24;j++){
           for(i=0;i<32;i++){
              k = attr[i+(j<<5)];
              j1 = zx(k&0x7);
              j2 = zx((k&0x38)>>3);
              jj = (k&0x40)?1:0;
              if(jj){j1|=8;j2|=8;}
              k = j1|(j2<<4);
              fprintf(f,"s%2.2d_%2.2d\tdb ",i,j);
              for(jj=0;jj<8;jj++)
              {
                 ii = i+(j<<8)+(jj<<5);
                 fprintf(f,"#%2.2X,",image[ii]);
              }
              fprintf(f,"#%2.2X\n",(unsigned char)k);
           }}
           fclose(f);
           strcat(str," saved!");
           _setcolor(7);
           OutText(MSGX,MSGY,str);
           delay(200);
        }
      }
      um.Show();
   }
 }
 free(buf);
 if(text!=copyr) free(text);
 _setvideomode(-1);
 return 1;
}
